------------------------------------------------------
-- GFWTooltip.lua
-- Version 3
-- Generic tooltip hooking
-- Credit where due: mostly based on GDI's Reagent Info (which in turn uses code from ItemsMatrix, LootLink, and Auctioneer)
-- Additional inspiration from Reagent Watch and EnhTooltip
------------------------------------------------------

------------------------------------------------------
-- External API
------------------------------------------------------

--[[ GFWTooltip_AddCallback(modName, callbackFunction)

Allows you to have a function called whenever a tooltip is displayed. 
The function is called with arguments (frame, name, link, source):

	frame: the instance of GameTooltip being shown, which you can modify via its API
	name: the first line of the tooltip, typically an item name.
	link: a hyperlink to the item being shown in the tooltip
	source: a non-localized string identifying why this tooltip is being shown 
			(e.g. "MERCHANT" for an item seen in the MerchantFrame)
	
Your callback function should return true if it modifies the tooltip, so we know not to modify the same tooltip twice.

Example:

	function MyTooltipCallback(frame, name, link, source)
		frame:AddLine("some stuff");
		return true;
	end
	function MyMod_OnLoad()
		GFWTooltip_AddCallback("MyMod", MyTooltipCallback);
	end
]]

local function addCallback(modName, callbackFunction)
	GFWTooltip_Callbacks[modName] = callbackFunction;
end

------------------------------------------------------
-- Initial setup
------------------------------------------------------

local GFWTOOLTIP_THIS_VERSION = 4;
if (GFWTooltip == nil) then
	GFWTooltip = {};
end
if (GFWTooltip_Callbacks == nil) then
	GFWTooltip_Callbacks = {};
end
if (GFWTooltip_OriginalFunctions == nil) then
	GFWTooltip_OriginalFunctions = {};
end
local G = GFWTooltip;
local Orig = GFWTooltip_OriginalFunctions;

------------------------------------------------------
-- Internal functions
------------------------------------------------------

local checkTooltip; -- If non-nil, check for appearance of GameTooltip for adding information
local checkTimer; -- Timer for frequency of tooltip checks
local gameToolTipOwner; -- The current owner of the GameTooltip
local currentTooltip; -- Current Tooltip frame
local itemsMatrixEnabled; -- Boolean to watch if ItemsMatrix is running

-- We call this at the bottom of this file to get things started. 
local function setupHookFunctions()

	-- Hooks for Blizzard's GameTooltip functions
	Orig.GameTooltip_SetLootItem = GameTooltip.SetLootItem;
	GameTooltip.SetLootItem = GFWTooltip_temp_hook_GameTooltip_SetLootItem;
	Orig.GameTooltip_SetQuestItem = GameTooltip.SetQuestItem;
	GameTooltip.SetQuestItem = GFWTooltip_temp_hook_GameTooltip_SetQuestItem;
	Orig.GameTooltip_SetQuestLogItem = GameTooltip.SetQuestLogItem;
	GameTooltip.SetQuestLogItem = GFWTooltip_temp_hook_GameTooltip_SetQuestLogItem;
	Orig.GameTooltip_SetInventoryItem = GameTooltip.SetInventoryItem;
	GameTooltip.SetInventoryItem = GFWTooltip_temp_hook_GameTooltip_SetInventoryItem;
	Orig.GameTooltip_SetMerchantItem = GameTooltip.SetMerchantItem;
	GameTooltip.SetMerchantItem = GFWTooltip_temp_hook_GameTooltip_SetMerchantItem;
	Orig.GameTooltip_SetCraftItem = GameTooltip.SetCraftItem;
	GameTooltip.SetCraftItem = GFWTooltip_temp_hook_GameTooltip_SetCraftItem;
	Orig.GameTooltip_SetTradeSkillItem = GameTooltip.SetTradeSkillItem;
	GameTooltip.SetTradeSkillItem = GFWTooltip_temp_hook_GameTooltip_SetTradeSkillItem;
	Orig.GameTooltip_SetAuctionSellItem = GameTooltip.SetAuctionSellItem;
	GameTooltip.SetAuctionSellItem = GFWTooltip_temp_hook_GameTooltip_SetAuctionSellItem;
	Orig.GameTooltip_OnHide = GameTooltip_OnHide;
	GameTooltip_OnHide = GFWTooltip_temp_hook_GameTooltip_OnHide;
	Orig.GameTooltip_SetOwner = GameTooltip.SetOwner;
	GameTooltip.SetOwner = GFWTooltip_temp_hook_GameTooltip_SetOwner;
	
	-- Install our own OnUpdate handler for GameTooltip
	--Orig.GameTooltip_OnUpdate = GameTooltip:GetScript("OnUpdate");
	--GameTooltip:SetScript("OnUpdate", GFWTooltip_temp_hook_GameTooltip_OnUpdate);

	-- Hooks for other Blizzard functions
	Orig.ContainerFrameItemButton_OnEnter = ContainerFrameItemButton_OnEnter;
	ContainerFrameItemButton_OnEnter = GFWTooltip_temp_hook_ContainerFrameItemButton_OnEnter;
	Orig.ContainerFrame_Update = ContainerFrame_Update;
	ContainerFrame_Update = GFWTooltip_temp_hook_ContainerFrame_Update;
	Orig.AuctionFrameItem_OnEnter = AuctionFrameItem_OnEnter;
	AuctionFrameItem_OnEnter = GFWTooltip_temp_hook_AuctionFrameItem_OnEnter;
	Orig.SetItemRef = SetItemRef;
	SetItemRef = GFWTooltip_temp_hook_SetItemRef;

	-- Hook ItemsMatrix's functions if they're available
	if (type(ItemsMatrix_AddExtraTooltipInfo) == "function") then
		itemsMatrixEnabled = true;
	end

	-- Hook AIOI's stuff if it's available and we're not running ItemsMatrix
	if(AllInOneInventory_Enabled and not itemsMatrixEnabled) then
		Orig.AllInOneInventory_ModifyItemTooltip = AllInOneInventory_ModifyItemTooltip;
		AllInOneInventory_ModifyItemTooltip = GFWTooltip_temp_hook_AllInOneInventory_ModifyItemTooltip;
	end

	-- Hook for MyInventory
	if (MyInventoryProfile and not itemsMatrixEnabled) then
		Orig.MyInventory_ContainerFrameItemButton_OnEnter = MyInventory_ContainerFrameItemButton_OnEnter;
		MyInventory_ContainerFrameItemButton_OnEnter = GFWTooltip_temp_hook_MyInventory_ContainerFrameItemButton_OnEnter;

		Orig.MyInventoryFrame_Update = MyInventoryFrame_Update;
		MyInventoryFrame_Update = GFWTooltip_temp_hook_MyInventoryFrame_Update;
	end

	-- Hook for the LootLink window's tooltip (not using LL's tooltip hook because we already cover everything else it does)
	Orig.LootLinkItemButton_OnEnter = LootLinkItemButton_OnEnter;
	LootLinkItemButton_OnEnter = GFWTooltip_temp_hook_LootLinkItemButton_OnEnter;
	
end

local function addTooltipInfo(frame, name, link, source)

	if (frame.gfwDone == 1) then return; end -- we've already been here

	local changedTooltip = false;
	for modName, callback in GFWTooltip_Callbacks do
		if (callback ~= nil and type(callback) == "function") then
			local modifiedInCallback = callback(frame, name, link, source);
			changedTooltip = changedTooltip or modifiedInCallback;
			if (modifiedInCallback) then
				--GFWUtils.Print("Ran tooltip callback for ".. modName..", tooltip modified.");
			else
				--GFWUtils.Print("Ran tooltip callback for ".. modName..", tooltip not modified.");
			end
		end
	end
	--GFWUtils.Print("Done with tooltip callbacks.");
	if (changedTooltip) then
		frame.gfwDone = 1;
		frame:Show();
	end

end

-- Checks the tooltip info for an item name.  If one is found and we haven't updated the tip already, process it.
local function checkTooltipInfo(frame, link, source)
	-- If we've already added our information, no need to do it again
	if ( not frame or frame.gfwDone or itemsMatrixEnabled) then
		return nil;
	end

	currentTooltip = frame;
	if (link == nil) then
		link = GFWTooltip_LastLink;
	end
	GFWTooltip_LastLink = link;

	if( frame:IsVisible() ) then
		local field = getglobal(frame:GetName().."TextLeft1");
		if( field and field:IsVisible() ) then
			local name = field:GetText();
			if( name and frame.gfwDone ~= 1) then
				addTooltipInfo(frame, name, link, source);
				return nil;
			end
		end
	end

	return 1;
end

local function nameFromLink(link)
	local name;
	if( not link ) then
		return nil;
	end
	for name in string.gfind(link, "|c%x+|Hitem:%d+:%d+:%d+:%d+|h%[(.-)%]|h|r") do
		return name;
	end
	return nil;
end

-- Calling this will allow us to automatically add information to tooltips when needed
local function autoInfoOn()
	lSuppressInfoAdd = nil;
end

-- Calling this will prevent us from automatically adding information to tooltips
local function autoInfoOff()
	lSuppressInfoAdd = 1;
end

local function findItemInBags(findName)
	for bag = 0, 4, 1 do
		size = GetContainerNumSlots(bag);
		if (size) then
			for slot = size, 1, -1 do
				local link = GetContainerItemLink(bag, slot);
				if (link) then
					local name = nameFromLink(link);
					if (itemName == findName) then
						return bag, slot;
					end
				end
			end
		end
	end
end


--------------------
-- Hook Functions --
--------------------

-- Ensures that we don't try and write to the tooltips too quickly.
function GFWTooltip_temp_hook_GameTooltip_OnUpdate(elapsed)
	if (Orig.GameTooltip_OnUpdate ~= nil) then
		Orig.GameTooltip_OnUpdate(elapsed)
	end
	checkTimer = checkTimer + elapsed;

	if( checkTimer >= 1 ) then
		if( checkTooltip ) then
			checkTooltip = checkTooltipInfo(currentTooltip);
		end
		checkTimer = 0;
	end
end

function GFWTooltip_temp_hook_GameTooltip_SetLootItem(this, slot)
	Orig.GameTooltip_SetLootItem(this, slot);

	local link = GetLootSlotLink(slot);
	local name = nameFromLink(link);
	addTooltipInfo(GameTooltip, name, link, "LOOT");
end

function GFWTooltip_temp_hook_GameTooltip_SetQuestItem(this, qtype, slot)
	Orig.GameTooltip_SetQuestItem(this, qtype, slot);

	local link = GetQuestItemLink(qtype, slot);
	local name = nameFromLink(link);
	addTooltipInfo(GameTooltip, name, link, "QUEST");
end

function GFWTooltip_temp_hook_GameTooltip_SetQuestLogItem(this, qtype, slot)
	Orig.GameTooltip_SetQuestLogItem(this, qtype, slot);

	local link = GetQuestLogItemLink(qtype, slot);
	local name = nameFromLink(link);
	addTooltipInfo(GameTooltip, name, link, "QUESTLOG");
end

function GFWTooltip_temp_hook_GameTooltip_SetMerchantItem(this, slot)
	Orig.GameTooltip_SetMerchantItem(this, slot);

	local link = GetMerchantItemLink(slot);
	local name = nameFromLink(link);
	addTooltipInfo(GameTooltip, name, link, "MERCHANT");
end

function GFWTooltip_temp_hook_AuctionFrameItem_OnEnter(type, index)
	Orig.AuctionFrameItem_OnEnter(type, index);

	local link = GetAuctionItemLink(type, index);
	local name = nameFromLink(link);
	addTooltipInfo(GameTooltip, name, link, "AUCTION");
end

function GFWTooltip_temp_hook_GameTooltip_SetInventoryItem(this, unit, slot)
	local hasItem, hasCooldown, repairCost = Orig.GameTooltip_SetInventoryItem(this, unit, slot);
	local link = GetInventoryItemLink(unit, slot);
	local name = nameFromLink(link);
	addTooltipInfo(GameTooltip, name, link, "INVENTORY");

	return hasItem, hasCooldown, repairCost;
end

function GFWTooltip_temp_hook_GameTooltip_SetCraftItem(this, skill, slot)
	Orig.GameTooltip_SetCraftItem(this, skill, slot);
	local link;
	if (slot) then
		link = GetCraftReagentItemLink(skill, slot);
		local name = nameFromLink(link);
		addTooltipInfo(GameTooltip, name, link, "CRAFT_REAGENT");
	else
		link = GetCraftItemLink(skill);
		local name = nameFromLink(link);
		addTooltipInfo(GameTooltip, name, link, "CRAFT_ITEM");
	end
end

function GFWTooltip_temp_hook_GameTooltip_SetTradeSkillItem(this, skill, slot)
	Orig.GameTooltip_SetTradeSkillItem(this, skill, slot);
	local link;
	if (slot) then
		link = GetTradeSkillReagentItemLink(skill, slot);
		local name = nameFromLink(link);
		addTooltipInfo(GameTooltip, name, link, "TRADESKILL_REAGENT");
	else
		link = GetTradeSkillItemLink(skill);
		local name = nameFromLink(link);
		addTooltipInfo(GameTooltip, name, link, "TRADESKILL_ITEM");
	end
end

function GFWTooltip_temp_hook_GameTooltip_SetAuctionSellItem(this)
	Orig.GameTooltip_SetAuctionSellItem(this);
    local name, texture, quantity, quality, canUse, price = GetAuctionSellItemInfo();
	if (name) then
		local bag, slot = findItemInBags(name);
		if (bag) then
			local link = GetContainerItemLink(bag, slot);
			addTooltipInfo(GameTooltip, name, link, "AUCTION_SELL");
		end
	end
end

function GFWTooltip_temp_hook_ContainerFrameItemButton_OnEnter()
	Orig.ContainerFrameItemButton_OnEnter();
	autoInfoOff();

	if (itemsMatrixEnabled) then
		autoInfoOn();
		return;
	end

	if (not InRepairMode()) then
		local frameID = this:GetParent():GetID();
		local buttonID = this:GetID();
		local link = GetContainerItemLink(frameID, buttonID);
		local name = nameFromLink(link);

		if( name ) then
			local texture, itemCount, locked, quality, readable = GetContainerItemInfo(frameID, buttonID);
			checkTooltipInfo(GameTooltip, link, "CONTAINER");
			GameTooltip:Show();
		end
	end
	autoInfoOn();
end

function GFWTooltip_temp_hook_ContainerFrame_Update(frame)
	Orig.ContainerFrame_Update(frame);
	autoInfoOff();

	if (itemsMatrixEnabled) then
		autoInfoOn();
		return;
	end

	if( not InRepairMode() and GameTooltip:IsVisible() ) then
		local frameID = frame:GetID();
		local frameName = frame:GetName();
		local iButton;
		for iButton = 1, frame.size do
			local button = getglobal(frameName.."Item"..iButton);
			if( GameTooltip:IsOwned(button) ) then
				local buttonID = button:GetID();
				local link = GetContainerItemLink(frameID, buttonID);
				local name = nameFromLink(link);

				if( name ) then
					local texture, itemCount, locked, quality, readable = GetContainerItemInfo(frameID, buttonID);
					checkTooltipInfo(GameTooltip, link, "CONTAINER");
					GameTooltip:Show();
				end
			end
		end
	end
	autoInfoOn();
end

function GFWTooltip_temp_hook_GameTooltip_OnHide()
	Orig.GameTooltip_OnHide();
	GameTooltip.gfwDone = nil;
	if ( currentTooltip ) then
		currentTooltip.gfwDone = nil;
		currentTooltip = nil;
	end
end

function GFWTooltip_temp_hook_GameTooltip_SetOwner(this, owner, anchor)
	Orig.GameTooltip_SetOwner(this, owner, anchor);
	gameToolTipOwner = owner;
end

function GFWTooltip_temp_hook_SetItemRef(link, text, button)
	Orig.SetItemRef(link, text, button);
	checkTooltip = checkTooltipInfo(ItemRefTooltip, link, "ITEMREF");
end

-- AIOI Hooks
function GFWTooltip_temp_hook_AllInOneInventory_ModifyItemTooltip(bag, slot, tooltipName)
	Orig.AllInOneInventory_ModifyItemTooltip(bag, slot, tooltipName);

	-- Verify AIOI is installed and running
	if(not AllInOneInventory_Enabled or itemsMatrixEnabled) then
		return;	
	end

	local tooltip = getglobal(tooltipName);
	if ( not tooltip ) then
		tooltip = getglobal("GameTooltip");
		tooltipName = "GameTooltip";
	end
	if ( not tooltip ) then
		return false;
	end

	if ( not InRepairMode() ) then
		local link = GetContainerItemLink(bag, slot);
		local name = nameFromLink(link);

		addTooltipInfo(GameTooltip, name, link, "AIOI");
	end
end

-- MyInventory Hooks
function GFWTooltip_temp_hook_MyInventory_ContainerFrameItemButton_OnEnter()
	Orig.MyInventory_ContainerFrameItemButton_OnEnter();

	if (GameTooltip:IsVisible() and not InRepairMode()) then
		local bag, slot = MyInventory_GetIdAsBagSlot(this:GetID());
		local _, stack = GetContainerItemInfo(bag, slot);

		local link = GetContainerItemLink(bag, slot);
		local name = nameFromLink(link);

		if(link and name) then
			addTooltipInfo(GameTooltip, name, link, "MYINV");
		end
	end
end

function GFWTooltip_temp_hook_MyInventoryFrame_Update(frame)
	Orig.MyInventoryFrame_Update(frame);

	local name = frame:GetName();
	for j=1, frame.size, 1 do
		local itemButton = getglobal(name.."Item"..j);

		if (GameTooltip:IsVisible() and GameTooltip:IsOwned(itemButton)) then
			tooltip = getglobal("GameTooltip");
			tooltipName = "GameTooltip";

			if ( not tooltip ) then
				return false;
			end

			local bag, slot = MyInventory_GetIdAsBagSlot(itemButton:GetID());
			local link = GetContainerItemLink(bag, slot);
			local name = nameFromLink(link);

			if (name ~= nil) then
				addTooltipInfo(GameTooltip, name, link, "MYINV");
			end
		end
	end
end

function GFWTooltip_temp_hook_LootLinkItemButton_OnEnter()
	Orig.LootLinkItemButton_OnEnter();
	local name = this:GetText();
	local itemLink = ItemLinks[name];
	local link;
	if (itemLink and itemLink.c and itemLink.i and LootLink_CheckItemServer(itemLink, LootLinkState.ServerNamesToIndices[GetCVar("realmName")])) then
		local item = string.gsub(itemLink.i, "(%d+):(%d+):(%d+):(%d+)", "%1:0:%3:%4");
		link = "|c"..itemLink.c.."|Hitem:"..item.."|h["..name.."]|h|r";
	end
	if (link) then
		addTooltipInfo(LootLinkTooltip, name, link, "LOOTLINK");
	end
end

------------------------------------------------------
-- load only if not already loaded
------------------------------------------------------

if (G.Version == nil or (tonumber(G.Version) ~= nil and G.Version < GFWTOOLTIP_THIS_VERSION)) then

	-- Initialize state variables
	checkTimer = 0; -- Timer for frequency of tooltip checks
	itemsMatrixEnabled = false; -- Boolean to watch if ItemsMatrix is running

	-- Export functions
	setupHookFunctions();
	G.AddCallback = addCallback;
	GFWTooltip_AddCallback = addCallback;

	-- Set version number
	G.Version = GFWTOOLTIP_THIS_VERSION;

	GFWUtils.Print("GFWTooltip loaded.");

end


